from uuid import uuid4
from pathlib import Path
from itertools import cycle
import pkg_resources

import dash_html_components as html
import dash_core_components as dcc

import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS

from dash.dependencies import Output, Input, State
from webviz_config import WebvizPluginABC
from webviz_config.webviz_assets import WEBVIZ_ASSETS
from everviz.data.load_csv.get_data import get_data


class ObjectivesPlot(WebvizPluginABC):
    """
    The ObjectivesPlot class implements a plugin for Webviz, for plotting the
    objective values generated during an Everest optimization.

    Generally there are multiple realizations for the objective function values
    at each batch. Additionally, there may be multiple objective functions that
    are optimized simultaneously. These can be plotted either all individually
    in the plot, or they are summarized by plotting the mean value together with
    a P10-P90 range.
    """

    def __init__(self, app, values_file, statistics_file):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.radio_id = f"dropdown-{uuid4()}"
        self.function_dropdown_id = f"dropdown-{uuid4()}"

        self.values_file = values_file
        self.statistics_file = statistics_file
        self.set_callbacks(app)

        ASSETS_DIR = Path(pkg_resources.resource_filename("everviz", "assets"))
        WEBVIZ_ASSETS.add(ASSETS_DIR / "axis_customization.css")

    def add_webvizstore(self):
        return [
            (get_data, [{"values_file": self.values_file}]),
            (get_data, [{"statistics_file": self.statistics_file}]),
        ]

    @property
    def layout(self):
        data = get_data(self.statistics_file)

        functions = data["function"].unique()
        function_dropdown_options = [{"label": i, "value": i} for i in functions]
        func_elements = [
            html.Label("Functions to plot:"),
            dcc.Dropdown(
                id=self.function_dropdown_id,
                options=function_dropdown_options,
                multi=True,
                value=[functions[0]],
            ),
        ]

        radio_options = ["Statistics", "Data"]
        radio_elements = [
            dcc.RadioItems(
                id=self.radio_id,
                options=[{"label": i, "value": i} for i in radio_options],
                value=radio_options[1],
                labelStyle={"display": "inline-block"},
                style={"margin-top": 20},
            )
        ]

        return html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            func_elements + radio_elements,
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [dcc.Graph(id=self.graph_id)],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(
            self.plugin_data_output,
            [self.plugin_data_requested],
            [State(self.radio_id, "value")],
        )
        def user_download_data(data_requested, radio_value):
            if data_requested:
                file_path = (
                    self.statistics_file
                    if radio_value == "Statistics"
                    else self.values_file
                )
                return WebvizPluginABC.plugin_data_compress(
                    [
                        {
                            "filename": Path(file_path).name,
                            "content": get_data(file_path).to_csv(),
                        }
                    ]
                )
            return ""

        @app.callback(
            Output(self.graph_id, "figure"),
            [Input(self.function_dropdown_id, "value"), Input(self.radio_id, "value"),],
        )
        def update_graph(func_list, radio_value):
            # # The key_list arguments is the list of functions to plot.
            if func_list is None:
                return {}

            # # Get the data, setting its index.
            if radio_value == "Statistics":
                data = get_data(self.statistics_file).set_index("function")
            else:
                data = get_data(self.values_file).set_index("function")

            # Make a cycle iterator over the plotly colors.
            colors = cycle(DEFAULT_PLOTLY_COLORS)

            traces = []
            for color, key in zip(colors, func_list):
                # Select all data belonging to the current key.
                func_data = data.xs(key)

                # Make the traces, with mean, P10 and P90, shading in between.
                if radio_value == "Statistics":
                    traces.extend(
                        [
                            go.Scatter(
                                y=func_data["P90"],
                                x=func_data["batch"],
                                mode="lines",
                                marker={"size": 10},
                                line={"color": color},
                                name=key + "(P90)",
                                showlegend=False,
                            ),
                            go.Scatter(
                                y=func_data["P10"],
                                x=func_data["batch"],
                                mode="lines",
                                line={"color": color},
                                marker={"size": 10},
                                fill="tonexty",
                                name=key + "(P10)",
                                showlegend=False,
                            ),
                            go.Scatter(
                                y=func_data["Mean"],
                                x=func_data["batch"],
                                mode="lines",
                                marker={"size": 10},
                                line={"color": color},
                                name=key,
                                showlegend=True,
                            ),
                        ]
                    )
                else:
                    traces.append(
                        go.Scatter(
                            y=func_data["value"],
                            x=func_data["batch"],
                            mode="markers",
                            marker={"size": 10},
                            name=key,
                            showlegend=True,
                        )
                    )

            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": "Batch"}, yaxis={"title": "Function Key Value"},
                ),
            }



import os
import pandas as pd
import numpy as np

import everviz

from everviz.pages.objectives import (
    _set_up_data_sources,
    _objective_values,
    _objective_statistics,
    _total_objective_values,
)

OBJECTIVES = [
    {"batch": 0, "realization": 1, "function": "f0", "value": 100, "simulation": 0},
    {"batch": 0, "realization": 1, "function": "f1", "value": 200, "simulation": 1},
    {"batch": 2, "realization": 2, "function": "f0", "value": 200, "simulation": 2},
    {"batch": 2, "realization": 2, "function": "f1", "value": 400, "simulation": 3},
    {"batch": 0, "realization": 3, "function": "f0", "value": 300, "simulation": 4},
    {"batch": 0, "realization": 3, "function": "f1", "value": 600, "simulation": 5},
    {"batch": 2, "realization": 4, "function": "f0", "value": 400, "simulation": 6},
    {"batch": 2, "realization": 4, "function": "f1", "value": 800, "simulation": 7},
]


def test_objective_values_data_frame():
    """Test for the correct layout and size of the objective values data frame."""
    objective_values = _objective_values(pd.DataFrame(OBJECTIVES))

    assert set(objective_values.columns) == {
        "batch",
        "function",
        "value",
        "simulation",
    }
    assert len(objective_values) == len(OBJECTIVES)
    assert set(objective_values["batch"]) == {obj["batch"] for obj in OBJECTIVES}
    assert set(objective_values["function"]) == {obj["function"] for obj in OBJECTIVES}
    assert set(objective_values["simulation"]) == {
        obj["simulation"] for obj in OBJECTIVES
    }


def test_objective_statistics_data_frame():
    """Test for the correct layout and size of the objective statistics data frame"""
    objective_statistics = _objective_statistics(pd.DataFrame(OBJECTIVES))

    assert set(objective_statistics.columns) == {
        "batch",
        "function",
        "Mean",
        "P10",
        "P90",
    }
    assert len(objective_statistics) == 2 * 2  # #batches * #functions
    assert set(objective_statistics["batch"]) == {obj["batch"] for obj in OBJECTIVES}
    assert set(objective_statistics["function"]) == {
        obj["function"] for obj in OBJECTIVES
    }


def test_objective_statistics_content():
    """Test for the correct content of the objectives statistics data frame."""
    objective_statistics = _objective_statistics(pd.DataFrame(OBJECTIVES))

    mean = []
    p10 = []
    p90 = []
    for func in ["f0", "f1"]:
        for batch in [0, 2]:
            values = [
                obj["value"]
                for obj in OBJECTIVES
                if obj["batch"] == batch and obj["function"] == func
            ]
            mean.append(np.mean(values))
            p10.append(np.quantile(values, q=0.1))
            p90.append(np.quantile(values, q=0.9))

    assert objective_statistics["Mean"].to_list() == mean
    assert objective_statistics["P10"].to_list() == p10
    assert objective_statistics["P90"].to_list() == p90


def test_set_up_sources(mocker, monkeypatch, tmpdir):
    mock_api = mocker.Mock()
    mock_api.output_folder = tmpdir
    os.mkdir(os.path.join(tmpdir, "everviz"))
    monkeypatch.setattr(
        everviz.pages.objectives,
        "_objective_values_from_api",
        mocker.Mock(return_value=pd.DataFrame(OBJECTIVES)),
    )
    total_data = pd.DataFrame(OBJECTIVES).drop(
        columns=["realization", "function", "simulation"]
    )
    monkeypatch.setattr(
        everviz.pages.objectives,
        "_total_objective_values_from_api",
        mocker.Mock(return_value=total_data),
    )
    data_source = _set_up_data_sources(mock_api)
    assert data_source.objective_values == tmpdir / "everviz" / "objective_values.csv"
    assert (
        data_source.objective_statistics
        == tmpdir / "everviz" / "objective_statistics.csv"
    )
    assert (
        data_source.total_objective_values
        == tmpdir / "everviz" / "total_objective_values.csv"
    )


def test_total_objective_values_data_frame():
    total_data = pd.DataFrame(OBJECTIVES).drop(
        columns=["realization", "function", "simulation"]
    )
    total_objective_values = _total_objective_values(total_data)

    assert set(total_objective_values.columns) == {"batch", "value"}
    assert len(total_objective_values) == len(OBJECTIVES)



import os
from everviz.util import get_everviz_folder, DEFAULT_CONFIG


def page_layout(api):
    everviz_path = get_everviz_folder(api)
    config_path = os.path.join(everviz_path, DEFAULT_CONFIG)
    return {
        "title": "Config editor",
        "content": [{"ConfigEditor": {"data_path": config_path,},},],
    }



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import pkg_resources
import sphinx_rtd_theme

# -- Project information -----------------------------------------------------

project = "Everviz"
copyright = "2020, Equinor"
author = "Equinor"


def get_version():
    # Retrieve the everviz installed version.
    everviz_version = pkg_resources.get_distribution(project).version
    print("everest version: {}".format(everviz_version))
    return everviz_version


# The short X.Y version
version = get_version()
# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_rtd_theme",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_sidebars = {
    "**": ["globaltoc.html", "relations.html", "sourcelink.html", "searchbox.html"]
}



import os
import time

import dash
import pytest
import dash_html_components as html
import pandas as pd
from PIL import Image

import everviz
from everviz.plugins import Crossplot
from everviz.pages import crossplot
from everviz.plugins.crossplot.layout.crossplot_layout import (
    get_sidebar_layout,
    _get_dropdown,
    _get_radio,
)


@pytest.mark.parametrize(
    "test_input,expected",
    [(["dropdown_val"], "dropdown_val"), (["1", "2"], "1"), (["1", "2"], "2")],
)
def test_crossplot_layout_dropdown(dash_duo, test_input, expected):
    app = dash.Dash(__name__)
    item_id = "id"
    layout = _get_dropdown(item_id=item_id, options=test_input)
    app.layout = html.Div(layout)
    dash_duo.start_server(app)

    dash_duo.select_dcc_dropdown("#{}".format(item_id), expected)
    assert dash_duo.find_element("#{}".format(item_id)).text.split()[0] == expected


@pytest.mark.parametrize(
    "valid_options", [[("OPT_1", "nr_1"), ("OPT_2", "nr_2")], ["OPT_1", "OPT_2"]]
)
@pytest.mark.parametrize("expected, placement", [("OPT_1", 1), ("OPT_2", 2)])
def test_crossplot_layout_radio(dash_duo, expected, placement, valid_options):

    app = dash.Dash(__name__)
    layout = _get_radio("radio_id", options=valid_options)
    app.layout = html.Div(layout)
    dash_duo.start_server(app)

    text = dash_duo.find_element(
        "#{} label:nth-child({})".format("radio_id", placement)
    ).text
    assert text == expected


def test_get_sidebar_layout(monkeypatch, mocker):
    radio_mock = mocker.Mock(return_value=[])
    dropdown_mock = mocker.Mock(return_value=[])
    monkeypatch.setattr(
        everviz.plugins.crossplot.layout.crossplot_layout, "_get_radio", radio_mock
    )
    monkeypatch.setattr(
        everviz.plugins.crossplot.layout.crossplot_layout,
        "_get_dropdown",
        dropdown_mock,
    )

    get_sidebar_layout(
        [("radio", {"item_id": "radio_id"}), ("dropdown", {"item_id": "dropdown_id"})]
    )

    radio_mock.assert_called_once_with(item_id="radio_id")
    dropdown_mock.assert_called_once_with(item_id="dropdown_id")


def test_crossplot_layout(dash_duo, monkeypatch, mocker, tmpdir, assert_equal_images):
    app = dash.Dash(__name__)
    mock_data = pd.DataFrame(data=[[1, 2, 3]], columns=["a", "b", "c"])
    monkeypatch.setattr(
        everviz.plugins.crossplot.crossplot,
        "get_data",
        mocker.Mock(return_value=mock_data),
    )
    monkeypatch.setattr(
        everviz.plugins.crossplot.crossplot.Crossplot, "set_callbacks", mocker.Mock()
    )

    plugin = Crossplot(mocker.Mock(), "data_path")
    layout = plugin.layout
    app.layout = layout
    dash_duo.start_server(app)

    _REFERENCE_IMAGES = {
        (1024, 768): "crossplot_layout_headless.png",
        (1016, 635): "crossplot_layout.png",
    }

    dash_duo.driver.set_window_size(1024, 768)
    with tmpdir.as_cwd():
        # Graph can be slow to load, so we need to ensure it has loaded.
        time.sleep(1)
        dash_duo.driver.save_screenshot("example_snapshot.png")
        snapshot = Image.open("example_snapshot.png")

    if not snapshot.size in _REFERENCE_IMAGES:
        pytest.skip("No reference image for layout size: {}".format(snapshot.size))

    reference_image = Image.open(
        os.path.join(
            everviz.__path__[0],
            "..",
            "test-data",
            "crossplot",
            _REFERENCE_IMAGES[snapshot.size],
        )
    )

    assert_equal_images(reference_image, snapshot, threshold=0.5)


def test_webviz_page_layout(mocker):
    api_mock = mocker.Mock()
    api_mock.everest_csv = "everest_export.csv"
    api_mock.control_names = ["x-0", "x-1"]
    result = crossplot.page_layout(api_mock)
    expected_page_layout = {
        "title": "Cross plots",
        "content": [
            crossplot._crossplot_doc,
            {"Crossplot": {"data_path": "everest_export.csv",},},
            crossplot._crossplot_indexed_doc,
            {"CrossplotIndexed": {"data_path": "everest_export.csv",}},
        ],
    }
    assert expected_page_layout == result


def test_webviz_page_layout_no_crossplot(mocker):
    api_mock = mocker.Mock()
    api_mock.everest_csv = "everest_export.csv"
    api_mock.control_names = ["x", "y"]
    result = crossplot.page_layout(api_mock)
    expected_page_layout = {
        "title": "Cross plots",
        "content": [
            crossplot._crossplot_doc,
            {"Crossplot": {"data_path": "everest_export.csv",},},
        ],
    }
    assert expected_page_layout == result


def test_webviz_page_layout_nothing(mocker):
    api_mock = mocker.Mock()
    api_mock.control_names = ["x-0", "x-1"]
    result = crossplot.page_layout(api_mock)
    api_mock.everest_csv = None
    result = crossplot.page_layout(api_mock)
    expected_page_layout = ""
    assert expected_page_layout == result



from everviz.plugins.crossplot.crossplot import Crossplot
from everviz.plugins.crossplot.crossplot_indexed import CrossplotIndexed
from everviz.plugins.summary_plot.summary_plot import SummaryPlot
from everviz.plugins.objectives_plot.objectives_plot import ObjectivesPlot
from everviz.plugins.objectives_plot.single_objectives_plot import SingleObjectivesPlot
from everviz.plugins.controls_plot.controls_plot import ControlsPlot
from everviz.plugins.config_editor.config_editor import ConfigEditor



import pandas as pd
import everviz
from everviz.plugins.crossplot.crossplot import Crossplot


def test_crossplot_callback(app, dash_duo, monkeypatch, mocker, caplog):
    mock_data = pd.DataFrame(data=[[1, 2, 3]], columns=["a", "b", "c"])
    monkeypatch.setattr(
        everviz.plugins.crossplot.crossplot,
        "get_data",
        mocker.Mock(return_value=mock_data),
    )

    plugin = Crossplot(app, "a_data_file")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    dash_duo.clear_input("#{}".format(plugin.keys_x_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"



import pytest
import dash
from PIL import ImageFilter, ImageChops


@pytest.fixture()
def assert_equal_images():
    def func(image_1, image_2, threshold=0.1):
        """
        In this function we check if two images are almost identical
         by blurring and calculating MSE between the images.
        """

        image_1, image_2 = [image.convert("L") for image in [image_1, image_2]]
        image_1, image_2 = [
            image.filter(ImageFilter.GaussianBlur(radius=4))
            for image in [image_1, image_2]
        ]

        pixel_pairs = zip(list(image_1.getdata()), list(image_2.getdata()))
        diff = sum([(pixel_1 - pixel_2) ** 2 for pixel_1, pixel_2 in pixel_pairs])

        error = diff / (image_1.size[0] * image_1.size[1])

        assert error < threshold, "Bounding box: {}".format(
            ImageChops.subtract(image_1, image_2).getbbox()
        )

    return func


@pytest.fixture()
def app():
    dash_app = dash.Dash(__name__)
    dash_app.config.suppress_callback_exceptions = True
    yield dash_app



import os
from pathlib import Path
from uuid import uuid4
import pkg_resources
import dash_html_components as html
import dash_core_components as dcc

from dash.dependencies import Input, Output
from webviz_config import WebvizPluginABC
from webviz_config.webviz_assets import WEBVIZ_ASSETS

from everviz.plugins.crossplot.layout.crossplot_layout import get_sidebar_layout
from everviz.plugins.crossplot.callback.crossplot_callback import (
    crossplot_update_graph,
    get_graph_line,
)
from everviz.data.load_csv.get_data import get_data


class Crossplot(WebvizPluginABC):
    """
    The Crossplot class implements a plugin for Webviz, which can plot any
    column from a CSV file against any other column. This is used by Everviz for
    plotting columns from CSV file exported by Everest .
    """

    def __init__(self, app, data_path, title="Crossplot"):
        super().__init__()
        self.title = title
        self.graph_id = f"graph-{uuid4()}"
        self.keys_x_id = f"dropdown-{uuid4()}"
        self.keys_y_id = f"dropdown-{uuid4()}"
        self.axis_type_x_id = f"radio-{uuid4()}"
        self.axis_type_y_id = f"radio-{uuid4()}"
        self.axis_opt_x_id = f"radio-{uuid4()}"
        self.axis_opt_y_id = f"radio-{uuid4()}"

        self.data_path = data_path
        self.set_callbacks(app)

        ASSETS_DIR = pkg_resources.resource_filename("everviz", os.path.join("assets"))

        WEBVIZ_ASSETS.add(Path(ASSETS_DIR) / "axis_customization.css")

    def add_webvizstore(self):
        return [(get_data, [{"data_path": self.data_path}])]

    @property
    def layout(self):
        axis_type = [(i, i) for i in ["Linear", "Log"]]
        axis_options = ["Normal", "Cumulative"]
        df = get_data(self.data_path)
        plot_keys = df.columns.unique()

        sidebar_configs = [
            (
                "radio",
                {
                    "title": "X-axis",
                    "item_id": self.axis_type_x_id,
                    "options": axis_type,
                },
            ),
            ("radio", {"item_id": self.axis_opt_x_id, "options": axis_options},),
            ("dropdown", {"item_id": self.keys_x_id, "options": plot_keys},),
            (
                "radio",
                {
                    "title": "Y-axis",
                    "item_id": self.axis_type_y_id,
                    "options": axis_type,
                },
            ),
            ("radio", {"item_id": self.axis_opt_y_id, "options": axis_options},),
            ("dropdown", {"item_id": self.keys_y_id, "options": plot_keys},),
        ]

        return html.Div(
            [
                html.H1(children=self.title, style={"textAlign": "center"},),
                html.Div(
                    [
                        html.Div(
                            [get_sidebar_layout(sidebar_configs)],
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [dcc.Graph(id=self.graph_id)],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return WebvizPluginABC.plugin_data_compress(
                    [
                        {
                            "filename": Path(self.data_path).name,
                            "content": get_data(self.data_path).to_csv(),
                        }
                    ]
                )
            return ""

        @app.callback(
            Output(self.graph_id, "figure"),
            [
                Input(self.keys_x_id, "value"),
                Input(self.keys_y_id, "value"),
                Input(self.axis_type_x_id, "value"),
                Input(self.axis_type_y_id, "value"),
                Input(self.axis_opt_x_id, "value"),
                Input(self.axis_opt_y_id, "value"),
            ],
        )
        def update_graph(
            xaxis_column_name,
            yaxis_column_name,
            xaxis_type,
            yaxis_type,
            axis_options_x_id,
            axis_options_y_id,
        ):
            if None in [xaxis_column_name, yaxis_column_name]:
                return {}
            df = get_data(self.data_path)
            x_data = df[xaxis_column_name]
            y_data = df[yaxis_column_name]
            data = [
                get_graph_line(x_data, y_data, axis_options_x_id, axis_options_y_id)
            ]
            return crossplot_update_graph(
                data, xaxis_column_name, yaxis_column_name, xaxis_type, yaxis_type
            )



import subprocess
import shutil
from everviz.util import get_everviz_folder
from everviz.config import setup_default_everviz_config
from everviz.log import setup_logger


# Mock the hookimpl decorator
# Allow everviz to be installed and have tests run without installing everest
def mock_impl(func):
    return func


try:
    from everest.plugins import hookimpl
except ImportError:
    hookimpl = mock_impl


@hookimpl
def visualize_data(api):
    everviz_folder = get_everviz_folder(api)
    logger = setup_logger(everviz_folder)

    config_path = setup_default_everviz_config(api)
    # The entry point of webviz is to call it from command line, and so do we.
    if shutil.which("webviz"):
        subprocess.call(["webviz", "build", config_path, "--theme", "equinor"])
    else:
        logger.error("Failed to find webviz")



import pandas as pd
from everviz.plugins.objectives_plot.objectives_plot import ObjectivesPlot
from everviz.pages.objectives import _objective_values, _objective_statistics


def test_objective_plot_callback(app, dash_duo, mocker, caplog):
    test_data = [
        {"batch": 0, "realization": 1, "function": "f0", "value": 100, "simulation": 0},
        {"batch": 0, "realization": 1, "function": "f1", "value": 200, "simulation": 1},
        {"batch": 2, "realization": 2, "function": "f0", "value": 200, "simulation": 2},
        {"batch": 2, "realization": 2, "function": "f1", "value": 400, "simulation": 3},
        {"batch": 0, "realization": 3, "function": "f0", "value": 300, "simulation": 4},
        {"batch": 0, "realization": 3, "function": "f1", "value": 600, "simulation": 5},
        {"batch": 2, "realization": 4, "function": "f0", "value": 400, "simulation": 6},
        {"batch": 2, "realization": 4, "function": "f1", "value": 800, "simulation": 7},
    ]

    def mock_get_data(data_type):
        if data_type == "values":
            result = _objective_values(pd.DataFrame(test_data))
        else:
            result = _objective_statistics(pd.DataFrame(test_data))
        return result

    mocker.patch(
        "everviz.plugins.objectives_plot.objectives_plot.get_data",
        side_effect=mock_get_data,
    )

    plugin = ObjectivesPlot(app, "values", "statistics")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    # Test statistics plot.
    dash_duo.select_dcc_dropdown("#{}".format(plugin.function_dropdown_id), "f1")

    # Test data plot.
    dash_duo.find_element("#{} label:nth-child({})".format(plugin.radio_id, 2)).click()

    # Clear the dropdown, which should not cause an error.
    dash_duo.clear_input("#{}".format(plugin.function_dropdown_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"



import os
from collections import namedtuple
from functools import partial

import pandas
import numpy

from everviz.log import get_logger


DataSources = namedtuple("DataSource", ["summary_values", "summary_statistics"])

logger = get_logger()

_SUMMARY_INDEX_COLUMNS = ["batch", "date", "simulation"]


def _summary_values(summary_values):
    # Sort by the index columns.
    sorted_values = (
        summary_values.set_index(_SUMMARY_INDEX_COLUMNS).sort_index().reset_index()
    )
    return sorted_values


def _summary_statistics(summary_values):
    # Find the key names.
    value_vars = [
        column
        for column in summary_values.columns
        if column not in _SUMMARY_INDEX_COLUMNS
    ]

    # Make rows out of the keys for statistics.
    reshaped_summary_values = pandas.melt(
        summary_values,
        id_vars=_SUMMARY_INDEX_COLUMNS,
        value_vars=value_vars,
        var_name="summary_key",
        value_name="value",
    )

    # Aggregate the values over the simulations, using pivot_table, keeping the
    # key, batch and date as the multi-index, calculating statistics of the
    # values over the simulations.
    summary_statistics = pandas.pivot_table(
        reshaped_summary_values,
        values="value",
        index=["summary_key", "batch", "date"],
        aggfunc=[
            numpy.mean,
            partial(numpy.quantile, q=0.1),
            partial(numpy.quantile, q=0.9),
        ],
    ).droplevel(1, axis="columns")
    summary_statistics.columns = ["mean", "P10", "P90"]

    # Sort the multi index, and reset them to columns.
    sorted_statistics = summary_statistics.sort_index().reset_index()

    return sorted_statistics


def _set_up_data_sources(api, keys=None):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")

    # Make a table which statistics over the simulations.
    summary_values = api.summary_values(keys=keys)

    if summary_values.empty:
        return None

    logger.info("Generating summary values plot")

    summary_values_file = os.path.join(everviz_path, "summary_values.csv")
    values = _summary_values(summary_values)
    values.to_csv(summary_values_file, index=False)
    logger.info(f"File created: {summary_values_file}")  # pylint: disable=W1203

    summary_statistics_file = os.path.join(everviz_path, "summary_statistics.csv")
    statistics = _summary_statistics(summary_values)
    statistics.to_csv(summary_statistics_file, index=False)
    logger.info(f"File created: {summary_statistics_file}")  # pylint: disable=W1203

    return DataSources(
        summary_values=summary_values_file, summary_statistics=summary_statistics_file,
    )


def page_layout(api):
    sources = _set_up_data_sources(api)

    if sources is None:
        return {}

    return {
        "title": "Summary Values",
        "content": [
            "## Summary values as a function of date",
            {
                "SummaryPlot": {
                    "values_file": sources.summary_values,
                    "statistics_file": sources.summary_statistics,
                    "xaxis": "date",
                },
            },
            "## Summary values as a function of batch",
            {
                "SummaryPlot": {
                    "values_file": sources.summary_values,
                    "statistics_file": sources.summary_statistics,
                    "xaxis": "batch",
                },
            },
        ],
    }



from pathlib import Path
from uuid import uuid4
from itertools import cycle
import pkg_resources

import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Output, Input, State

import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS

from webviz_config import WebvizPluginABC
from webviz_config.webviz_assets import WEBVIZ_ASSETS
from everviz.data.load_csv.get_data import get_data


class SummaryPlot(WebvizPluginABC):
    """
    The SummaryPlot class implements a plugin for Webviz, for plotting summary
    key values generated by the eclipse or flow simulators during an Everest
    optimization.

    Summary values can be plotted as a function of time, or as a function of the
    batch number. Generally there are multiple realizations at each batch or
    date. These can be plotted either all individually in the plot, or they are
    summarized by plotting the mean value together with a P10-P90 range.
    """

    def __init__(self, app, values_file, statistics_file, xaxis="date"):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.key_dropdown_id = f"dropdown-{uuid4()}"
        self.xaxis_dropdown_id = f"dropdown-{uuid4()}"
        self.radio_id = f"radio-{uuid4()}"

        self.values_file = values_file
        self.statistics_file = statistics_file
        self.xaxis = xaxis

        self.set_callbacks(app)

        ASSETS_DIR = pkg_resources.resource_filename("everviz", "assets")
        WEBVIZ_ASSETS.add(Path(ASSETS_DIR) / "axis_customization.css")

    def add_webvizstore(self):
        return [
            (get_data, [{"values_file": self.values_file}]),
            (get_data, [{"statistics_file": self.statistics_file}]),
        ]

    @property
    def layout(self):
        radio_options = ["Statistics", "Data"]
        data = get_data(self.statistics_file)
        key_dropdown_options = [
            {"label": i, "value": i} for i in list(data["summary_key"].unique())
        ]
        xaxis_dropdown_options = [
            {"label": i, "value": i}
            for i in list(data["batch" if self.xaxis == "date" else "date"].unique())
        ]
        xaxis_dropdown_title = (
            "Batches to plot" if self.xaxis == "date" else "Dates to plot"
        )

        # Keywords dropdown.
        keyword_elements = [
            html.Label("Keywords to plot:"),
            dcc.Dropdown(
                id=self.key_dropdown_id, options=key_dropdown_options, multi=True,
            ),
        ]

        # X-axis dropdown.
        xaxis_elements = [
            html.Label(xaxis_dropdown_title, style={"margin-top": 24}),
            dcc.Dropdown(
                id=self.xaxis_dropdown_id,
                options=xaxis_dropdown_options,
                multi=True,
                value=[xaxis_dropdown_options[0]["value"]],
            ),
        ]

        # Radio for switching between data and statistics.
        radio_elements = [
            dcc.RadioItems(
                id=self.radio_id,
                options=[{"label": i, "value": i} for i in radio_options],
                value=radio_options[0],
                labelStyle={"display": "inline-block"},
                style={"margin-top": 20},
            ),
        ]
        return html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            keyword_elements + xaxis_elements + radio_elements,
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [dcc.Graph(id=self.graph_id)],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(
            self.plugin_data_output,
            [self.plugin_data_requested],
            [State(self.radio_id, "value")],
        )
        def user_download_data(data_requested, radio_value):
            if data_requested:
                if radio_value == "Statistics":
                    data = get_data(self.statistics_file).set_index(
                        ["summary_key", "batch", "date"]
                    )
                    file_path = self.statistics_file
                else:
                    data = get_data(self.values_file).set_index(["batch", "date"])
                    file_path = self.values_file

                return WebvizPluginABC.plugin_data_compress(
                    [{"filename": Path(file_path).name, "content": data.to_csv(),}]
                )
            return ""

        @app.callback(
            Output(self.graph_id, "figure"),
            [
                Input(self.key_dropdown_id, "value"),
                Input(self.xaxis_dropdown_id, "value"),
                Input(self.radio_id, "value"),
            ],
        )
        def update_graph(key_list, line_list, radio_value):
            # The key_list arguments is the list of keys to plot. The line_list
            # argument is a list of batches, or a list of dates to plot for
            # those keys.
            if key_list is None or line_list is None:
                return {}

            # Get the data, setting its index.
            if radio_value == "Statistics":
                data = get_data(self.statistics_file).set_index(
                    ["summary_key", "batch", "date"]
                )
            else:
                data = get_data(self.values_file).set_index(["batch", "date"])

            # Make a cycle iterator over the plotly colors.
            colors = cycle(DEFAULT_PLOTLY_COLORS)

            # Choose between dates or batches for different lines.
            line_key = "batch" if self.xaxis == "date" else "date"

            traces = []
            for color, key in zip(colors, key_list):
                # Select all data belonging to the current key.
                if radio_value == "Statistics":
                    key_data = data.xs(key, level="summary_key", drop_level=True)
                else:
                    key_data = data[key]

                for line in line_list:
                    # Select all rows belonging the current batch or date.
                    line_data = key_data.xs(
                        line, level=line_key, drop_level=True
                    ).reset_index()

                    # Set the name of the plotted line.
                    name = key
                    if len(line_list) > 1:
                        name += f", {line_key}:{line}"

                    # Make the traces, with mean, P10 and P90, shading in between.
                    if radio_value == "Statistics":
                        traces.extend(
                            [
                                go.Scatter(
                                    y=line_data["P90"],
                                    x=line_data[self.xaxis],
                                    mode="lines",
                                    marker={"size": 10},
                                    line={"color": color},
                                    name=name + "(P90)",
                                    showlegend=False,
                                ),
                                go.Scatter(
                                    y=line_data["P10"],
                                    x=line_data[self.xaxis],
                                    mode="lines",
                                    line={"color": color},
                                    marker={"size": 10},
                                    fill="tonexty",
                                    name=name + "(P10)",
                                    showlegend=False,
                                ),
                                go.Scatter(
                                    y=line_data["mean"],
                                    x=line_data[self.xaxis],
                                    mode="lines",
                                    marker={"size": 10},
                                    line={"color": color},
                                    name=name,
                                    showlegend=True,
                                ),
                            ]
                        )
                    else:
                        traces.append(
                            go.Scatter(
                                y=line_data[key],
                                x=line_data[self.xaxis],
                                mode="markers",
                                marker={"size": 10},
                                name=name,
                                showlegend=True,
                            )
                        )

            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": self.xaxis.capitalize()},
                    yaxis={"title": "Summary Key Value"},
                ),
            }



from datetime import datetime
import pandas as pd
from everviz.plugins.summary_plot.summary_plot import SummaryPlot
from everviz.pages.summary_values import _summary_values, _summary_statistics


def test_summary_plot_callback(app, dash_duo, mocker, caplog):
    test_data = {
        "simulation": range(12),
        "batch": [0] * 6 + [1] * 6,
        "date": [
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
        ],
        "key1": range(1, 13),
        "key2": range(10, 130, 10),
    }

    def mock_get_data(data_type):
        if data_type == "values":
            result = _summary_values(pd.DataFrame(test_data))
        else:
            result = _summary_statistics(pd.DataFrame(test_data))
        return result

    mocker.patch(
        "everviz.plugins.summary_plot.summary_plot.get_data", side_effect=mock_get_data,
    )

    plugin = SummaryPlot(app, "values", "statistics")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    # Test statistics plot.
    dash_duo.select_dcc_dropdown("#{}".format(plugin.key_dropdown_id), "key1")

    # Test data plot.
    dash_duo.find_element("#{} label:nth-child({})".format(plugin.radio_id, 2)).click()

    # Clear the axes dropdowns, which should not cause an error.
    #
    # NOTE: if you try to do this before the previous click action on a radio
    # button, you get an error, because the the dropdown stays open and overlaps
    # with the radio buttons.
    dash_duo.clear_input("#{}".format(plugin.xaxis_dropdown_id))
    dash_duo.clear_input("#{}".format(plugin.key_dropdown_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"



import os
from collections import defaultdict

DEFAULT_CONFIG = "everviz_webviz_config.yml"


def get_everviz_folder(api):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")
    if not os.path.exists(everviz_path):
        os.makedirs(everviz_path)
    return everviz_path


def identify_indexed_controls(controls):
    base_names = defaultdict(list)
    for control in controls:
        try:
            base_name, index = control.split("-")
            if index.isnumeric():
                base_names[base_name].append(int(index))
        except ValueError:
            continue

    return base_names



from uuid import uuid4
from pathlib import Path
import pkg_resources

import dash_html_components as html
import dash_core_components as dcc

import plotly.graph_objs as go

from dash.dependencies import Output, Input
from webviz_config import WebvizPluginABC
from webviz_config.webviz_assets import WEBVIZ_ASSETS
from everviz.data.load_csv.get_data import get_data


class ControlsPlot(WebvizPluginABC):
    """
    The ControlsPlot class implements a plugin for Webviz, for plotting the
    control values generated during an Everest optimization.
    """

    def __init__(self, app, csv_file):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.controls_dropdown_id = f"dropdown-{uuid4()}"

        self.csv_file = csv_file
        self.set_callbacks(app)

        ASSETS_DIR = Path(pkg_resources.resource_filename("everviz", "assets"))
        WEBVIZ_ASSETS.add(ASSETS_DIR / "axis_customization.css")

    def add_webvizstore(self):
        return [(get_data, [{"csv_file": self.csv_file}])]

    @property
    def layout(self):
        data = get_data(self.csv_file)

        controls = data["control"].unique()
        control_dropdown_options = [{"label": i, "value": i} for i in controls]
        control_elements = [
            html.Label("Controls to plot:"),
            dcc.Dropdown(
                id=self.controls_dropdown_id,
                options=control_dropdown_options,
                multi=True,
                value=[controls[0]],
            ),
        ]

        return html.Div(
            [
                html.Div(
                    [
                        html.Div(
                            control_elements,
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [dcc.Graph(id=self.graph_id)],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return WebvizPluginABC.plugin_data_compress(
                    [
                        {
                            "filename": Path(self.csv_file).name,
                            "content": get_data(self.csv_file).to_csv(),
                        }
                    ]
                )
            return ""

        @app.callback(
            Output(self.graph_id, "figure"),
            [Input(self.controls_dropdown_id, "value")],
        )
        def update_graph(control_list):
            # # The key_list arguments is the list of functions to plot.
            if control_list is None:
                return {}

            # # Get the data, setting its index.
            data = get_data(self.csv_file).set_index("control")

            traces = []
            for control in control_list:
                control_data = data.xs(control)
                traces.append(
                    go.Scatter(
                        y=control_data["value"],
                        x=control_data["batch"],
                        mode="lines",
                        name=control,
                        showlegend=True,
                    ),
                )
            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": "Batch"}, yaxis={"title": "Control Value"},
                ),
            }



import pandas as pd

import pytest
from everviz.pages.objectives import (
    _total_objective_values_from_api,
    _single_objective_title,
)

TEST_DATA = [
    {"batch": 0, "value": 100, "accepted": True},
    {"batch": 1, "value": 200, "accepted": True},
    {"batch": 2, "value": 200, "accepted": False},
    {"batch": 3, "value": 400, "accepted": True},
    {"batch": 4, "value": 300, "accepted": False},
    {"batch": 5, "value": 600, "accepted": False},
    {"batch": 6, "value": 400, "accepted": True},
    {"batch": 7, "value": 800, "accepted": True},
]


@pytest.fixture
def api_mock(mocker):
    api = mocker.Mock()
    single_objectives = [
        {"batch": data["batch"], "value": data["value"]} for data in TEST_DATA
    ]
    api.single_objective_values = single_objectives
    accepted_batches = [data["batch"] for data in TEST_DATA if data["accepted"]]
    api.accepted_batches = accepted_batches
    return api


def test_total_objective_values_from_api(api_mock):
    result = _total_objective_values_from_api(api_mock)
    expected = pd.DataFrame(TEST_DATA)
    assert expected.equals(result)


def test_single_objective_title(api_mock):
    api_mock.objective_function_names = ["npv"]
    result = _single_objective_title(api_mock)
    expected = "Objective function"
    assert result == expected

    api_mock.objective_function_names = ["npv", "rf"]
    result = _single_objective_title(api_mock)
    expected = "Weighted objective function"
    assert result == expected



import os
import pytest
from everviz.util import get_everviz_folder


@pytest.fixture
def mocked_api(mocker):
    api_mock = mocker.Mock()
    api_mock.output_folder = "everest_output"
    return api_mock


def test_get_everviz_folder(tmpdir, mocked_api):
    everest_folder = "everest_output"
    expected_everviz_path = os.path.join(everest_folder, "everviz")
    with tmpdir.as_cwd():
        assert not os.path.exists(expected_everviz_path)
        everviz_path = get_everviz_folder(mocked_api)
        assert os.path.exists(everviz_path)
        assert expected_everviz_path == everviz_path



import pytest
from everviz.plugins.crossplot.callback.crossplot_indexed_dropdown import (
    dropdown_callback,
)


@pytest.mark.parametrize(
    "test_control, expected",
    [
        (
            None,
            [
                {"label": "a", "value": "a"},
                {"label": "b", "value": "b"},
                {"label": "c", "value": "c"},
            ],
        ),
        ("a", [{"label": "a", "value": "a"}, {"label": "b", "value": "b"}]),
        ("c", [{"label": "c", "value": "c"}]),
    ],
)
def test_crossplot_dropdown_callback(test_control, expected):
    indexed_control = {"a": [1, 2, 3], "b": [1, 2, 3], "c": [1]}
    result = dropdown_callback(test_control, indexed_control)
    assert result == expected



import time
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from everviz.plugins.config_editor.config_editor import ConfigEditor

everviz_config_1 = """title: Everest Optimization Report Test
pages:
- title: Page1
  content:
  - '## Testing config editor'
- title: Page2
  content:
  - '## Testing config editor'"""

everviz_config_2 = """title: Everest Optimization Report Test
pages:
- title: Page1
  content:
  - '## Testing config editor'"""


config_editor_page = """
- title: Config editor
  content:
  - ConfigEditor:
      data_path: everviz_config.yml"""


def write_text_area(driver, css_selector, text):
    """Simulate key press to clear the input."""
    elem = driver.find_element_by_css_selector(css_selector)
    (
        ActionChains(driver)
        .move_to_element(elem)
        .pause(0.2)
        .click(elem)
        .key_down(Keys.CONTROL)
        .send_keys(Keys.END)
        .key_down(Keys.SHIFT)
        .send_keys(Keys.HOME)
        .key_up(Keys.SHIFT)
        .key_up(Keys.CONTROL)
        .send_keys(Keys.DELETE)
        .send_keys(text)
    ).perform()


def test_config_editor_callback(app, dash_duo, caplog, tmpdir):
    with tmpdir.as_cwd():
        config_file = "everviz_config.yml"
        with open(config_file, "w") as f:
            f.write(everviz_config_1)

        plugin = ConfigEditor(app, config_file)
        app.layout = plugin.layout
        dash_duo.start_server(app)

        dash_duo.wait_for_text_to_equal(f"#{plugin.md_area}", everviz_config_1, 1)
        dash_duo.wait_for_contains_text(f"#{plugin.text_area}", everviz_config_1, 1)

        dash_duo.find_element(f"#{plugin.btn_edit}").click()
        time.sleep(1)
        write_text_area(dash_duo.driver, f"#{plugin.text_area}", "TEST")
        dash_duo.wait_for_contains_text(f"#{plugin.text_area}", "TEST", 1)
        dash_duo.find_element(f"#{plugin.btn_cancel}").click()
        dash_duo.wait_for_text_to_equal(f"#{plugin.md_area}", everviz_config_1, 1)
        dash_duo.wait_for_contains_text(f"#{plugin.text_area}", everviz_config_1, 1)

        dash_duo.find_element(f"#{plugin.btn_edit}").click()
        time.sleep(1)
        write_text_area(dash_duo.driver, f"#{plugin.text_area}", everviz_config_2)
        dash_duo.wait_for_contains_text(f"#{plugin.text_area}", everviz_config_2, 1)
        assert not plugin.default_conf_path.exists()

        dash_duo.find_element(f"#{plugin.btn_edit}").click()
        dash_duo.wait_for_text_to_equal(
            f"#{plugin.md_area}", everviz_config_2 + config_editor_page, 1
        )
        dash_duo.wait_for_contains_text(
            f"#{plugin.text_area}", everviz_config_2 + config_editor_page, 1
        )
        assert plugin.default_conf_path.exists()

        dash_duo.find_element(f"#{plugin.btn_edit}").click()
        time.sleep(1)
        write_text_area(dash_duo.driver, f"#{plugin.text_area}", "TEST")
        dash_duo.wait_for_contains_text(f"#{plugin.text_area}", "TEST", 1)
        dash_duo.find_element(f"#{plugin.btn_edit}").click()
        dash_duo.wait_for_text_to_equal(
            f"#{plugin.md_area}", everviz_config_2 + config_editor_page, 1
        )
        dash_duo.wait_for_contains_text(
            f"#{plugin.text_area}", everviz_config_2 + config_editor_page, 1
        )

        dash_duo.find_element(f"#{plugin.btn_reset}").click()
        dash_duo.wait_for_text_to_equal(f"#{plugin.md_area}", everviz_config_1, 1)
        dash_duo.wait_for_contains_text(f"#{plugin.text_area}", everviz_config_1, 1)

        for record in caplog.records:
            assert record.levelname != "ERROR"



import dash_core_components as dcc
import dash_html_components as html


def get_sidebar_layout(configurations):
    _FUNCTION_MAP = {"radio": _get_radio, "dropdown": _get_dropdown}

    divs = []
    for comp_name, args in configurations:
        divs += _FUNCTION_MAP[comp_name](**args)
    return html.Div(divs)


def _get_radio(item_id, title=None, options=None):
    if not options:
        options = []
    try:
        options = [{"label": i, "value": j} for i, j in options]
    except ValueError:
        options = [{"label": i, "value": i} for i in options]
    return [
        title,
        dcc.RadioItems(
            id=item_id,
            options=options,
            value=options[0]["value"] if len(options) > 0 else None,
            labelStyle={"display": "inline-block"},
        ),
    ]


def _get_dropdown(item_id, title=None, options=None, multi=False):
    if options is None:
        options = []
    return [
        title,
        dcc.Dropdown(
            id=item_id,
            options=[{"label": i, "value": i} for i in options],
            placeholder="Select a label",
            value=options[0] if len(options) > 0 else [],
            multi=multi,
        ),
    ]



from uuid import uuid4
from pathlib import Path
import pkg_resources

import dash_html_components as html
import dash_core_components as dcc

import plotly.graph_objs as go

from dash.dependencies import Output, Input
from webviz_config import WebvizPluginABC
from webviz_config.webviz_assets import WEBVIZ_ASSETS
from everviz.data.load_csv.get_data import get_data


class SingleObjectivesPlot(WebvizPluginABC):
    """
    This webviz plugin is designed to display the single objective function
    calculated as the weighted sum of the objective functions from a multi
    objective everest optimisation.
    """

    def __init__(self, app, csv_file):
        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.div_id = f"div-{uuid4()}"

        self.csv_file = csv_file
        self.set_callbacks(app)

        ASSETS_DIR = Path(pkg_resources.resource_filename("everviz", "assets"))
        WEBVIZ_ASSETS.add(ASSETS_DIR / "axis_customization.css")

    def add_webvizstore(self):
        return [
            (get_data, [{"csv_file": self.csv_file}]),
        ]

    @property
    def layout(self):
        return html.Div(
            [dcc.Graph(id=self.graph_id)],
            id=self.div_id,
            style={"width": "69%", "display": "inline-block"},
        )

    def set_callbacks(self, app):
        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return WebvizPluginABC.plugin_data_compress(
                    [
                        {
                            "filename": Path(self.csv_file).name,
                            "content": get_data(self.csv_file).to_csv(),
                        }
                    ]
                )
            return ""

        @app.callback(
            Output(self.graph_id, "figure"), [Input(self.div_id, "children")],
        )
        def update_graph(_):
            data = get_data(self.csv_file)
            accepted_data = data[data.accepted == True]  # pylint: disable=C0121
            not_accepted_data = data[data.accepted == False]  # pylint: disable=C0121
            traces = [
                go.Scatter(
                    y=accepted_data.value,
                    x=accepted_data.batch,
                    mode="lines+markers",
                    marker={"color": "blue", "size": 10},
                    name="accepted",
                ),
                go.Scatter(
                    y=not_accepted_data.value,
                    x=not_accepted_data.batch,
                    mode="markers",
                    marker={"color": "red", "size": 9},
                    name="not accepted",
                ),
            ]

            return {
                "data": traces,
                "layout": dict(
                    xaxis={"title": "Batch"},
                    yaxis={"title": "Objection function value"},
                ),
            }



import pytest

from everviz import everest_hooks


def test_everest_hooks_registered():
    hook_manager = pytest.importorskip("everest.plugins.hook_manager")

    pm = hook_manager.EverestPluginManager()
    try:
        pm.register(everest_hooks)
    except ValueError as err:
        if not str(err).startswith("Plugin already registered"):
            raise err

    assert any(
        [
            hook.plugin_name.startswith("everviz")
            for hook in pm.hook.visualize_data.get_hookimpls()
        ]
    )



import base64
import io
import os
from pathlib import Path
from uuid import uuid4
from dash import callback_context
from dash.dependencies import Input, Output, State
import dash_html_components as html
import dash_core_components as dcc
from webviz_config import WebvizPluginABC
import yaml


class ConfigEditor(WebvizPluginABC):
    """
    Webviz plugin designed to allow the user to edit a everviz config file.
     Plugin supports the following functionality, allow the user to:
     * Edit the content of the config file and save the changes
       only if the result is a valid everviz config file
     * Rest to the initial version of the config file.
     * Upload content from a local file.
    """

    def __init__(self, app, data_path, title="Configuration"):
        super().__init__()
        self.title = title
        self.data_path = data_path
        self.text_area = f"txt_area-{uuid4()}"
        self.md_area = f"md_area-{uuid4()}"
        self.editor_buttons = f"editor_buttons-{uuid4()}"
        self.btn_edit = f"btn_edit-{uuid4()}"
        self.btn_cancel = f"btn_cancel-{uuid4()}"
        self.btn_upload = f"btn_upload-{uuid4()}"
        self.btn_reset = f"btn_reset-{uuid4()}"
        self.config_component = f"config_component-{uuid4()}"
        self.editing = False
        self.default_conf_path = Path(self.data_path).parent / "default"
        self.default_conf = yaml.safe_load(self.get_config_str())
        self.set_callbacks(app)

    def get_config_str(self):
        with open(self.data_path, "r") as f:
            return f.read()

    @staticmethod
    def write_config(config, path):
        with open(path, "w") as f:
            yaml.safe_dump(config, f, default_flow_style=False, sort_keys=False)

    def update_everviz_config(self, new_config_text):
        new_config = yaml.safe_load(new_config_text)
        old_config = yaml.safe_load(self.get_config_str())
        if isinstance(new_config, dict) and old_config != new_config:
            pages = new_config.get("pages", [])
            if len(pages) > 0:
                # Ensure the config editor page is always part of the config
                conf_ed_content = {
                    "ConfigEditor": {"data_path": self.data_path,},
                }
                # Check if the ConfigEditor plugin is part of the configuration
                found = next(
                    (page for page in pages if page["content"] == [conf_ed_content]),
                    None,
                )
                if not found:
                    pages.append(
                        {"title": "Config editor", "content": [conf_ed_content]}
                    )
                if not self.default_conf_path.exists():
                    self.write_config(self.default_conf, self.default_conf_path)
                self.write_config(new_config, self.data_path)

    def render_text_component(self):
        config_str = self.get_config_str()
        if self.editing:
            style = {"width": "100%", "height": 800}
            return [dcc.Textarea(id=self.text_area, value=config_str, style=style,)]

        md_content = f"""```
{config_str}
```
"""
        return [
            dcc.Textarea(id=self.text_area, value=config_str, hidden=not self.editing,),
            dcc.Markdown(id=self.md_area, children=md_content,),
        ]

    def render_buttons(self):
        if not self.editing:
            return [
                html.Button("Edit", id=self.btn_edit),
                # Had to wrap the button in a div because the hidden option was not
                # working on the button element
                html.Div([html.Button("Cancel", id=self.btn_cancel),], hidden=True),
                html.Div(
                    dcc.Upload(html.Button("Upload File"), id=self.btn_upload),
                    style={"display": "inline-block",},
                ),
                html.Div(
                    [html.Button("Reset", id=self.btn_reset),],
                    hidden=not self.default_conf_path.exists(),
                ),
            ]
        return [
            html.Button("Save", id=self.btn_edit),
            html.Button("Cancel", id=self.btn_cancel),
            html.Div(html.Button("Reset", id=self.btn_reset,), hidden=True),
            dcc.Upload(children=None, id=self.btn_upload),
        ]

    @property
    def layout(self):
        return html.Div(
            [
                html.H1(children=self.title, style={"textAlign": "center"},),
                html.Div(
                    [
                        html.Div(
                            id=self.config_component,
                            children=self.render_text_component(),
                            style={
                                "width": "100%",
                                "height": "100%",
                                "display": "inline-block",
                            },
                        ),
                        html.Div(
                            id=self.editor_buttons,
                            children=self.render_buttons(),
                            style={"textAlign": "center"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(
            [
                Output(self.config_component, "children"),
                Output(self.editor_buttons, "children"),
            ],
            [
                Input(self.btn_cancel, "n_clicks"),
                Input(self.btn_edit, "n_clicks"),
                Input(self.btn_reset, "n_clicks"),
                Input(self.btn_upload, "contents"),
            ],
            [State(self.text_area, "value")],
        )
        def update_layout(
            cancel_clicks, edit_clicks, reset_clicks, upload_content, value
        ):
            changed_id = [p["prop_id"] for p in callback_context.triggered][0]
            if edit_clicks and self.btn_edit in changed_id:
                if self.editing:
                    self.update_everviz_config(value)
                self.editing = not self.editing
            elif cancel_clicks and self.btn_cancel in changed_id:
                if self.editing:
                    self.editing = not self.editing
            elif reset_clicks and self.btn_reset in changed_id:
                os.rename(self.default_conf_path, self.data_path)

            elif upload_content is not None:
                content_type, content_string = upload_content.split(",")
                decoded = base64.b64decode(content_string).decode("utf-8")
                new_config = "".join(io.StringIO(decoded).readlines())

                self.update_everviz_config(new_config)

            return self.render_text_component(), self.render_buttons()



import os
from collections import namedtuple
from functools import partial
import pandas as pd
import numpy as np

from everviz.log import get_logger

DataSources = namedtuple(
    "DataSource", ["objective_values", "objective_statistics", "total_objective_values"]
)

logger = get_logger()


def _objective_values_from_api(api):
    return pd.DataFrame(api.objective_values)


def _total_objective_values_from_api(api):
    single_objectives = api.single_objective_values
    accepted_batches = api.accepted_batches
    for objective in single_objectives:
        objective.update({"accepted": objective["batch"] in accepted_batches})
    return pd.DataFrame(single_objectives)


def _objective_values(data):
    # Sort by the index columns.
    data = data.drop(columns=["realization"])
    sorted_values = (
        data.set_index(["function", "batch", "simulation"]).sort_index().reset_index()
    )
    return sorted_values


def _total_objective_values(data):
    # Sort by the index columns.
    sorted_values = data.set_index("batch").sort_index().reset_index()
    return sorted_values


def _objective_statistics(data):
    # Aggregate the values over the simulations, using pivot_table, keeping the
    # batch and function as the multi-index, calculating statistics of the
    # values over the simulations.
    statistics = pd.pivot_table(
        data,
        values="value",
        index=["function", "batch"],
        aggfunc=[np.mean, partial(np.quantile, q=0.1), partial(np.quantile, q=0.9),],
    )
    statistics.columns = ["Mean", "P10", "P90"]

    # Sort the multi index, and reset them to columns.
    sorted_statistics = statistics.sort_index().reset_index()

    return sorted_statistics


def _set_up_data_sources(api):
    everest_folder = api.output_folder
    everviz_path = os.path.join(everest_folder, "everviz")

    logger.info("Generating objective values plot")
    data = _objective_values_from_api(api)

    objective_values_file = os.path.join(everviz_path, "objective_values.csv")
    values = _objective_values(data)
    values.to_csv(objective_values_file, index=False)
    logger.info(f"File created: {objective_values_file}")  # pylint: disable=W1203

    objective_statistics_file = os.path.join(everviz_path, "objective_statistics.csv")
    statistics = _objective_statistics(data)
    statistics.to_csv(objective_statistics_file, index=False)
    logger.info(f"File created: {objective_statistics_file}")  # pylint: disable=W1203

    logger.info("Generating total objective values plot")
    data = _total_objective_values_from_api(api)

    total_objective_values_file = os.path.join(
        everviz_path, "total_objective_values.csv"
    )
    values = _total_objective_values(data)
    values.to_csv(total_objective_values_file, index=False)
    logger.info(f"File created: {total_objective_values_file}")  # pylint: disable=W1203

    return DataSources(
        objective_values=objective_values_file,
        objective_statistics=objective_statistics_file,
        total_objective_values=total_objective_values_file,
    )


def _single_objective_title(api):
    nr_function = len(api.objective_function_names)
    if nr_function > 1:
        return "Weighted objective function"
    return "Objective function"


def page_layout(api):
    sources = _set_up_data_sources(api)
    return {
        "title": "Objectives",
        "content": [
            "## Objective function values",
            {
                "ObjectivesPlot": {
                    "values_file": sources.objective_values,
                    "statistics_file": sources.objective_statistics,
                },
            },
            f"## {_single_objective_title(api)}",
            {"SingleObjectivesPlot": {"csv_file": sources.total_objective_values,},},
        ],
    }



import pandas as pd
from everviz.plugins.controls_plot.controls_plot import ControlsPlot


def test_control_plot_callback(app, dash_duo, mocker, caplog):
    test_data = [
        {"control": "c1", "batch": 0, "value": 0},
        {"control": "c2", "batch": 0, "value": 0},
        {"control": "c1", "batch": 1, "value": 42},
        {"control": "c2", "batch": 1, "value": 24},
    ]

    mocker.patch(
        "everviz.plugins.controls_plot.controls_plot.get_data",
        return_value=pd.DataFrame(test_data),
    )

    plugin = ControlsPlot(app, "dummy")
    app.layout = plugin.layout
    dash_duo.start_server(app)

    # Test adding a control.
    dash_duo.select_dcc_dropdown("#{}".format(plugin.controls_dropdown_id), "c2")

    # Clear the dropdown, which should not cause an error. Call twice to
    # simulate two clear key presses.
    dash_duo.clear_input("#{}".format(plugin.controls_dropdown_id))
    dash_duo.clear_input("#{}".format(plugin.controls_dropdown_id))

    for record in caplog.records:
        assert record.levelname != "ERROR"



#!/usr/bin/env python
from os import path
from setuptools import setup, find_packages

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="everviz",
    packages=find_packages(exclude=["tests", "test-data"]),
    package_data={"everviz": ["assets/axis_customization.css"]},
    description="Visualization for Everest",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Software Innovation Bergen, Equinor ASA",
    license="AGPL-3.0",
    url="https://github.com/equinor/everviz",
    setup_requires=["setuptools_scm"],
    install_requires=[
        "pyyaml",
        "pandas",
        "numpy",
        "dash",
        "webviz-config",
        "plotly",
        "flask",
    ],
    test_suite="tests",
    use_scm_version={"write_to": "everviz/version.py"},
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points={
        "webviz_config_plugins": [
            "Crossplot = everviz.plugins:Crossplot",
            "CrossplotIndexed = everviz.plugins:CrossplotIndexed",
            "SummaryPlot = everviz.plugins:SummaryPlot",
            "ObjectivesPlot = everviz.plugins:ObjectivesPlot",
            "SingleObjectivesPlot = everviz.plugins:SingleObjectivesPlot",
            "ControlsPlot = everviz.plugins:ControlsPlot",
            "ConfigEditor = everviz.plugins:ConfigEditor",
        ],
        "everest": ["everviz = everviz.everest_hooks",],
    },
)



import logging
import sys
import os


def _formatter():
    return logging.Formatter("%(asctime)s — %(name)s — %(levelname)s — %(message)s")


def _console_handler():
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(_formatter())
    return handler


def _file_handler(path):
    if not os.path.isdir(path):
        os.makedirs(path)
    file_handler = logging.FileHandler(os.path.join(path, "everviz.log"))
    file_handler.setFormatter(_formatter())
    return file_handler


def setup_logger(path):
    logger = logging.getLogger("everviz")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(_console_handler())
    logger.addHandler(_file_handler(path))
    logger.propagate = False
    return logger


def get_logger():
    return logging.getLogger("everviz")



import numpy as np
import plotly.graph_objs as go


def crossplot_update_graph(
    data, xaxis_column_name, yaxis_column_name, xaxis_type, yaxis_type
):
    return {
        "data": data,
        "layout": dict(
            xaxis={"title": xaxis_column_name, "type": xaxis_type.lower(),},
            yaxis={"title": yaxis_column_name, "type": yaxis_type.lower(),},
        ),
    }


def get_graph_line(
    x_data,
    y_data,
    xaxis_option,
    yaxis_option,
    mode="markers",
    line_shape=None,
    name=None,
):
    return go.Scatter(
        y=y_data if yaxis_option != "Cumulative" else np.cumsum(y_data),
        x=x_data if xaxis_option != "Cumulative" else np.cumsum(x_data),
        mode=mode,
        marker={"size": 10},
        line_shape=line_shape,
        name=name,
    )



import os
from pathlib import Path
from uuid import uuid4

import pkg_resources
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Output, Input
from webviz_config import WebvizPluginABC
from webviz_config.webviz_assets import WEBVIZ_ASSETS
from everviz.data.load_csv.get_data import get_data
from everviz.plugins.crossplot.callback.crossplot_callback import (
    get_graph_line,
    crossplot_update_graph,
)
from everviz.util import identify_indexed_controls
from everviz.plugins.crossplot.callback.crossplot_indexed_dropdown import (
    dropdown_callback,
)
from everviz.plugins.crossplot.layout.crossplot_layout import get_sidebar_layout


class CrossplotIndexed(WebvizPluginABC):
    """
    The CrossplotIndexed class implements a plugin for Webviz, which can plot
    any indexed control from the data exported by Everest against any other
    indexed control.
    """

    def __init__(self, app, data_path, title="Indexed Crossplot"):
        super().__init__()
        self.title = title
        self.graph_id = f"graph-{uuid4()}"
        self.dropdown_x_id = f"dropdown-{uuid4()}"
        self.dropdown_y_id = f"dropdown-{uuid4()}"
        self.axis_type_x_id = f"radio-{uuid4()}"
        self.axis_type_y_id = f"radio-{uuid4()}"
        self.axis_options_x_id = f"radio-{uuid4()}"
        self.axis_options_y_id = f"radio-{uuid4()}"
        self.dropdown_realization_id = f"dropdown-{uuid4()}"
        self.interpolation_id = f"interpolation-{uuid4()}"

        self.data_path = data_path
        self.set_callbacks(app)

        ASSETS_DIR = pkg_resources.resource_filename("everviz", os.path.join("assets"))

        WEBVIZ_ASSETS.add(Path(ASSETS_DIR) / "axis_customization.css")

    def add_webvizstore(self):
        return [(get_data, [{"data_path": self.data_path}])]

    @property
    def layout(self):
        df = get_data(self.data_path)
        dropdown_options = list(identify_indexed_controls(df.columns.unique()).keys())
        realizations = df.index.values
        axis_type = [(i, i) for i in ["Linear", "Log"]]
        axis_options = ["Normal", "Cumulative"]
        interp_types = [
            ("None", None),
            ("Linear", "linear"),
            ("High value", "hv"),
            ("Low value", "vh"),
        ]

        side_bar_config = [
            (
                "radio",
                {
                    "title": "X-axis",
                    "item_id": self.axis_type_x_id,
                    "options": axis_type,
                },
            ),
            ("radio", {"item_id": self.axis_options_x_id, "options": axis_options},),
            ("dropdown", {"item_id": self.dropdown_x_id, "options": dropdown_options},),
            (
                "radio",
                {
                    "title": "Y-axis",
                    "item_id": self.axis_type_y_id,
                    "options": axis_type,
                },
            ),
            ("radio", {"item_id": self.axis_options_y_id, "options": axis_options},),
            ("dropdown", {"item_id": self.dropdown_y_id, "options": dropdown_options},),
            (
                "radio",
                {
                    "title": "Interpolation type",
                    "item_id": self.interpolation_id,
                    "options": interp_types,
                },
            ),
            (
                "dropdown",
                {
                    "item_id": self.dropdown_realization_id,
                    "options": realizations,
                    "multi": True,
                },
            ),
        ]

        return html.Div(
            [
                html.H1(children=self.title, style={"textAlign": "center"}),
                html.Div(
                    [
                        html.Div(
                            [get_sidebar_layout(side_bar_config)],
                            style={
                                "width": "29%",
                                "display": "inline-block",
                                "vertical-align": "top",
                            },
                        ),
                        html.Div(
                            [dcc.Graph(id=self.graph_id)],
                            style={"width": "69%", "display": "inline-block"},
                        ),
                    ]
                ),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(
            Output(self.dropdown_y_id, "options"),
            [Input(self.dropdown_x_id, "value"),],
        )
        def update_dropwdown_y(selected_control):
            df = get_data(self.data_path)
            indexed_controls = identify_indexed_controls(df.columns.unique())
            return dropdown_callback(selected_control, indexed_controls)

        @app.callback(
            Output(self.dropdown_y_id, "value"), [Input(self.dropdown_y_id, "options")],
        )
        def set_y_val_dropdown(available_options):
            try:
                return available_options[0]["value"]
            except IndexError:
                return None

        @app.callback(self.plugin_data_output, [self.plugin_data_requested])
        def user_download_data(data_requested):
            if data_requested:
                return WebvizPluginABC.plugin_data_compress(
                    [
                        {
                            "filename": Path(self.data_path).name,
                            "content": get_data(self.data_path).to_csv(),
                        }
                    ]
                )
            return ""

        @app.callback(
            Output(self.graph_id, "figure"),
            [
                Input(self.dropdown_x_id, "value"),
                Input(self.dropdown_y_id, "value"),
                Input(self.dropdown_realization_id, "value"),
                Input(self.axis_type_x_id, "value"),
                Input(self.axis_type_y_id, "value"),
                Input(self.interpolation_id, "value"),
                Input(self.axis_options_x_id, "value"),
                Input(self.axis_options_y_id, "value"),
            ],
        )
        def update_graph(
            xaxis_name,
            yaxis_name,
            realization_nr,
            xaxis_type,
            yaxis_type,
            interpolation_type,
            xaxis_opt,
            yaxis_opt,
        ):
            if None in [xaxis_name, yaxis_name]:
                return {}

            if isinstance(realization_nr, int):
                realization_nr = [realization_nr]

            df = get_data(self.data_path)
            traces = []
            for realization in realization_nr:
                x_data = df.filter(like=xaxis_name).loc[realization]
                y_data = df.filter(like=yaxis_name).loc[realization]
                mode = "markers" if not interpolation_type else "markers+lines"
                traces.append(
                    get_graph_line(
                        x_data,
                        y_data,
                        xaxis_opt,
                        yaxis_opt,
                        mode=mode,
                        line_shape=interpolation_type,
                        name="Realization: {}".format(realization_nr),
                    )
                )
            return crossplot_update_graph(
                traces, xaxis_name, yaxis_name, xaxis_type, yaxis_type
            )
